from dataclasses import dataclass
from enum import StrEnum
from importlib import resources
from pathlib import Path
from typing import Iterable

import yaml
import yaml_include
from cactus_test_definitions.csipaus import CSIPAusVersion
from cactus_test_definitions.errors import TestProcedureDefinitionError
from cactus_test_definitions.schema import UniqueKeyLoader
from cactus_test_definitions.server.actions import Action, validate_action_parameters
from cactus_test_definitions.server.checks import Check, validate_check_parameters
from dataclass_wizard import YAMLWizard


class TestProcedureId(StrEnum):
    """The set of all available test ID's

    This should be kept in sync with the current set of test procedures loaded from the procedures directory"""

    __test__ = False  # Prevent pytest from picking up this class
    S_ALL_01 = "S-ALL-01"
    S_ALL_02 = "S-ALL-02"
    S_ALL_03 = "S-ALL-03"
    S_ALL_04 = "S-ALL-04"
    S_ALL_05 = "S-ALL-05"
    S_OPT_01 = "S-OPT-01"
    S_OPT_02 = "S-OPT-02"
    S_OPT_03 = "S-OPT-03"
    S_OPT_04 = "S-OPT-04"


class ClientType(StrEnum):
    DEVICE = "device"  # This is a direct device client - i.e. the cert will match a SPECIFIC EndDevice
    AGGREGATOR = "aggregator"  # This is an aggregator client - i.e. the cert can manage MANY EndDevices


@dataclass
class RequiredClient:
    """A RequiredClient is a way for a test to assert that it needs a specific client type or set of clients. The id
    will be used internally within a test to reference a specific client"""

    id: str  # How this client will be referred to within the step's of the test
    client_type: ClientType | None = None  # If set - the client type that is required


@dataclass
class Step:
    """A step is an action for a client to execute and then a series of checks to validate the results. If the action
    raises an exception OR any of the checks fail, this step will marked as failed and the test will be aborted.

    Actions might represent a single operation or they may represent a series of polls/checks over a period of time.
    """

    id: str  # Descriptive identifier for this step (must be unique)
    action: Action  # The action to execute when the step starts
    client: str | None = None  # The RequiredClient.id that will execute this step. If None - use the 0th client.
    use_client_context: str | None = (
        None  # Specify to allow a request to execute with clientX using the context of clientY
    )
    checks: list[Check] | None = None  # The checks (if any) to execute AFTER action completes to determine success
    instructions: list[str] | None = None  # Text to display while this step executes

    repeat_until_pass: bool = False  # If True - failing checks will cause this step to re-execute until successful


@dataclass
class Preconditions:
    """Preconditions are a way of setting up the test / server before the test begins.

    Instructions are out-of-band information to show until the preconditions are all met.

    If any checks are required - they will be polled regularly until ALL pass. For each check poll, a discovery
    will be run to ensure data is available.
    """

    required_clients: list[RequiredClient]  # What client(s) need to be supplied to run this test procedure


@dataclass
class TestProcedure:
    """Top level object for collecting everything relevant to a single TestProcedure"""

    __test__ = False  # Prevent pytest from picking up this class
    description: str  # Metadata from test definitions
    category: str  # Metadata from test definitions
    classes: list[str]  # Metadata from test definitions
    target_versions: list[CSIPAusVersion]  # What version(s) of csip-aus is this test targeting?
    preconditions: Preconditions
    steps: list[Step]  # What behavior will the test procedure be evaluating?


@dataclass
class TestProcedures(YAMLWizard):
    """Represents a collection of CSIP-AUS server test procedure descriptions/specifications

    By sub-classing the YAMLWizard mixin, we get access to the class method `from_yaml`
    which we can use to create an instances of `TestProcedures`.
    """

    __test__ = False  # Prevent pytest from picking up this class

    description: str
    version: str
    test_procedures: dict[str, TestProcedure]

    def validate(self):
        for tp_name, tp in self.test_procedures.items():

            # Check preconditions
            if not tp.preconditions.required_clients:
                raise TestProcedureDefinitionError(
                    f"{tp_name} has no RequiredClients element. At least 1 entry required"
                )
            required_clients_by_id = dict(((rc.id, rc) for rc in tp.preconditions.required_clients))

            for step in tp.steps:
                validate_action_parameters(tp_name, step.id, step.action)

                # Validate step checks
                if step.checks:
                    for check in step.checks:
                        validate_check_parameters(tp_name, check)

                # Ensure client exists
                if step.client is not None and step.client not in required_clients_by_id:
                    raise TestProcedureDefinitionError(
                        f"{tp_name} reference client {step.client} that isn't listed in RequiredClients."
                    )


class TestProcedureConfig:
    __test__ = False  # Prevent pytest from picking up this class

    @staticmethod
    def from_yamlfile(path: Path, skip_validation: bool = False) -> TestProcedures:
        """Converts a yaml file given by 'path' into a 'TestProcedures' instance.

        Supports parts of the TestProcedures instance being described by external
        YAML files. These are referenced in the parent yaml file using the `!include` directive.

        skip_validation: If True, the test_procedures.validate() call will not be made

        Example:

            Description: CSIP-AUS Client Test Procedures
            Version: 0.1
            TestProcedures:
              ALL-01: !include ALL-01.yaml
        """
        with open(path, "r") as f:
            yaml_contents = f.read()

        # Modifies the pyyaml's load method to support references to external yaml files
        # through the `!include` directive.
        yaml.add_constructor(
            "!include", constructor=yaml_include.Constructor(base_dir=path.parent), Loader=UniqueKeyLoader
        )

        # ...because we are using YAMLWizard we need to supply a decoder and a Loader to
        # use this modified version.
        test_procedures: TestProcedures = TestProcedures.from_yaml(
            yaml_contents,
            decoder=yaml.load,  # type: ignore
            Loader=UniqueKeyLoader,
        )

        if not skip_validation:
            test_procedures.validate()

        return test_procedures

    @staticmethod
    def from_resource() -> TestProcedures:
        yaml_resource = resources.files("cactus_test_definitions.server.procedures") / "test-procedures.yaml"
        with resources.as_file(yaml_resource) as yaml_file:
            return TestProcedureConfig.from_yamlfile(path=yaml_file)

    @staticmethod
    def available_tests() -> Iterable[str]:
        test_procedures: TestProcedures = TestProcedureConfig.from_resource()
        return test_procedures.test_procedures.keys()



from dataclasses import dataclass
from enum import StrEnum
from importlib import resources
from pathlib import Path
from typing import Iterable

import yaml
import yaml_include
from cactus_test_definitions.client.actions import Action, validate_action_parameters
from cactus_test_definitions.client.checks import Check, validate_check_parameters
from cactus_test_definitions.client.events import Event, validate_event_parameters
from cactus_test_definitions.csipaus import CSIPAusVersion
from cactus_test_definitions.errors import TestProcedureDefinitionError
from cactus_test_definitions.schema import UniqueKeyLoader
from dataclass_wizard import YAMLWizard


class TestProcedureId(StrEnum):
    """The set of all available test ID's

    This should be kept in sync with the current set of client test procedures loaded from the procedures directory"""

    __test__ = False  # Prevent pytest from picking up this class
    ALL_01 = "ALL-01"
    ALL_02 = "ALL-02"
    ALL_03 = "ALL-03"
    ALL_03_REJ = "ALL-03-REJ"
    ALL_04 = "ALL-04"
    ALL_05 = "ALL-05"
    ALL_06 = "ALL-06"
    ALL_07 = "ALL-07"
    ALL_08 = "ALL-08"
    ALL_09 = "ALL-09"
    ALL_10 = "ALL-10"
    ALL_11 = "ALL-11"
    ALL_12 = "ALL-12"
    ALL_13 = "ALL-13"
    ALL_14 = "ALL-14"
    ALL_15 = "ALL-15"
    ALL_16 = "ALL-16"
    ALL_17 = "ALL-17"
    ALL_18 = "ALL-18"
    ALL_19 = "ALL-19"
    ALL_20 = "ALL-20"
    ALL_21 = "ALL-21"
    ALL_22 = "ALL-22"
    ALL_23 = "ALL-23"
    ALL_24 = "ALL-24"
    ALL_25 = "ALL-25"
    ALL_26 = "ALL-26"
    ALL_27 = "ALL-27"
    ALL_28 = "ALL-28"
    ALL_29 = "ALL-29"
    ALL_30 = "ALL-30"
    DRA_01 = "DRA-01"
    DRA_02 = "DRA-02"
    DRD_01 = "DRD-01"
    DRL_01 = "DRL-01"
    GEN_01 = "GEN-01"
    GEN_02 = "GEN-02"
    GEN_03 = "GEN-03"
    GEN_04 = "GEN-04"
    GEN_05 = "GEN-05"
    GEN_06 = "GEN-06"
    GEN_07 = "GEN-07"
    GEN_08 = "GEN-08"
    GEN_09 = "GEN-09"
    GEN_10 = "GEN-10"
    GEN_11 = "GEN-11"
    GEN_12 = "GEN-12"
    GEN_13 = "GEN-13"
    LOA_01 = "LOA-01"
    LOA_02 = "LOA-02"
    LOA_03 = "LOA-03"
    LOA_04 = "LOA-04"
    LOA_05 = "LOA-05"
    LOA_06 = "LOA-06"
    LOA_07 = "LOA-07"
    LOA_08 = "LOA-08"
    LOA_09 = "LOA-09"
    LOA_10 = "LOA-10"
    LOA_11 = "LOA-11"
    LOA_12 = "LOA-12"
    LOA_13 = "LOA-13"
    MUL_01 = "MUL-01"
    MUL_02 = "MUL-02"
    MUL_03 = "MUL-03"
    OPT_1_IN_BAND = "OPT-1-IN-BAND"
    OPT_1_OUT_OF_BAND = "OPT-1-OUT-OF-BAND"

    # Storage extension
    BES_01 = "BES-01"
    BES_02 = "BES-02"
    BES_03 = "BES-03"
    BES_04 = "BES-04"


@dataclass
class Step:
    """A step is a part of the test procedure that waits for some form of event before running a set of actions.

    It's common for a step to activate other "steps" so that the state of the active test procedure can "evolve" in
    response to client behaviour

    Instructions are out-of-band operations that need performing during the step
    e.g. disconnect DER from grid, disable power consumption etc.
    """

    event: Event  # The event to act as a trigger
    actions: list[Action]  # The actions to execute when the trigger is met
    instructions: list[str] | None = None


@dataclass
class Preconditions:
    """Preconditions are run during the "initialization" state that precedes the start of a test. They typically
    allow for the setup of the test.

    Checks are also included to prevent a client from starting a test before they have correctly met preconditions

    Instructions are out-of-band operations that need performing at the start of the test procedure
    e.g. attach a load etc.

    If immediate_start is set to True - the "initialization" step will be progressed through immediately so that the
    client has no opportunity to interact with the server in this state. Any actions will still be executed. Do NOT
    utilise immediate_start with precondition checks.
    """

    init_actions: list[Action] | None = None  # To be executed as the runner starts (before anything can occur)
    immediate_start: bool = False  # If True - a test execution will have NO "pre-start" phase.
    actions: list[Action] | None = None  # To be executed as the test case "starts" (usually on request of client)
    checks: list[Check] | None = None  # Will prevent move from "init" state to "started" state of a test if any fail
    instructions: list[str] | None = None


@dataclass
class Criteria:
    """Criteria represent the final pass/fail analysis run after a TestProcedure completion. They can consider both
    the final state of the test system as well as the interactions that happened while it was running"""

    checks: list[Check] | None = None  # These should be run at test procedure finalization to determine pass/fail


@dataclass
class TestProcedure:
    """Top level object for collecting everything relevant to a single TestProcedure"""

    __test__ = False  # Prevent pytest from picking up this class
    description: str  # Metadata from test definitions
    category: str  # Metadata from test definitions
    classes: list[str]  # Metadata from test definitions
    target_versions: list[CSIPAusVersion]  # What version(s) of csip-aus is this test targeting?
    steps: dict[str, Step]
    preconditions: Preconditions | None = None  # These execute during "init" and setup the test for a valid start state
    criteria: Criteria | None = None  # How will success/failure of this procedure be determined?


@dataclass
class TestProcedures(YAMLWizard):
    """Represents a collection of CSIP-AUS client test procedure descriptions/specifications

    By sub-classing the YAMLWizard mixin, we get access to the class method `from_yaml`
    which we can use to create an instances of `TestProcedures`.
    """

    __test__ = False  # Prevent pytest from picking up this class

    description: str
    version: str
    test_procedures: dict[str, TestProcedure]

    def _do_action_validate(self, procedure: TestProcedure, procedure_name: str, location: str, action: Action):
        """Handles the full validation of an action's definition for a parent procedure.

        procedure: The parent TestProcedure for action
        procedure_name: The name of procedure (used for labelling errors)
        location: Where in procedure can you find action? (used for labelling errors)
        action: The action to validate

        raises TestProcedureDefinitionError on failure
        """
        validate_action_parameters(procedure_name, location, action)

        # Provide additional "action specific" validation
        match action.type:
            case "enable-steps" | "remove-steps":
                for step_name in action.parameters["steps"]:
                    if step_name not in procedure.steps.keys():
                        raise TestProcedureDefinitionError(
                            f"{procedure_name}.{location}. Refers to unknown step '{step_name}'."
                        )

    def _validate_actions(self):
        """Validate actions of test procedure steps / preconditions

        Ensure,
        - action has the correct parameters
        - if parameters refer to steps then those steps are defined for the test procedure
        """

        for test_procedure_name, test_procedure in self.test_procedures.items():
            # Validate actions in the preconditions
            if test_procedure.preconditions:
                if test_procedure.preconditions.actions:
                    for action in test_procedure.preconditions.actions:
                        self._do_action_validate(test_procedure, test_procedure_name, "Precondition", action)

                if test_procedure.preconditions.init_actions:
                    for action in test_procedure.preconditions.init_actions:
                        self._do_action_validate(test_procedure, test_procedure_name, "Precondition", action)

            # Validate actions that exist on steps
            for step_name, step in test_procedure.steps.items():
                for action in step.actions:
                    self._do_action_validate(test_procedure, test_procedure_name, step_name, action)

    def _validate_checks(self):
        """Validate checks of test procedures

        Ensure,
        - check has the correct parameters
        """

        for test_procedure_name, test_procedure in self.test_procedures.items():
            if test_procedure.criteria and test_procedure.criteria.checks:
                for check in test_procedure.criteria.checks:
                    validate_check_parameters(test_procedure_name, check)

    def _validate_events(self):
        """Validate events of test procedure steps

        Ensure,
        - event has the correct parameters
        """

        for test_procedure_name, test_procedure in self.test_procedures.items():
            for step_name, step in test_procedure.steps.items():
                validate_event_parameters(test_procedure_name, step_name, step.event)

    def validate(self):
        self._validate_actions()
        self._validate_checks()
        self._validate_events()


class TestProcedureConfig:
    __test__ = False  # Prevent pytest from picking up this class

    @staticmethod
    def from_yamlfile(path: Path, skip_validation: bool = False) -> TestProcedures:
        """Converts a yaml file given by 'path' into a 'TestProcedures' instance.

        Supports parts of the TestProcedures instance being described by external
        YAML files. These are referenced in the parent yaml file using the `!include` directive.

        skip_validation: If True, the test_procedures.validate() call will not be made

        Example:

            Description: CSIP-AUS Client Test Procedures
            Version: 0.1
            TestProcedures:
              ALL-01: !include ALL-01.yaml
        """
        with open(path, "r") as f:
            yaml_contents = f.read()

        # Modifies the pyyaml's load method to support references to external yaml files
        # through the `!include` directive.
        yaml.add_constructor(
            "!include", constructor=yaml_include.Constructor(base_dir=path.parent), Loader=UniqueKeyLoader
        )

        # ...because we are using YAMLWizard we need to supply a decoder and a Loader to
        # use this modified version.
        test_procedures: TestProcedures = TestProcedures.from_yaml(
            yaml_contents,
            decoder=yaml.load,  # type: ignore
            Loader=UniqueKeyLoader,
        )

        if not skip_validation:
            test_procedures.validate()

        return test_procedures

    @staticmethod
    def from_resource() -> TestProcedures:
        yaml_resource = resources.files("cactus_test_definitions.client.procedures") / "test-procedures.yaml"
        with resources.as_file(yaml_resource) as yaml_file:
            return TestProcedureConfig.from_yamlfile(path=yaml_file)

    @staticmethod
    def available_tests() -> Iterable[str]:
        test_procedures: TestProcedures = TestProcedureConfig.from_resource()
        return test_procedures.test_procedures.keys()



from dataclasses import dataclass
from typing import Any

from cactus_test_definitions.client.checks import Check
from cactus_test_definitions.errors import TestProcedureDefinitionError
from cactus_test_definitions.parameters import (
    ParameterSchema,
    ParameterType,
    validate_parameters,
)
from cactus_test_definitions.variable_expressions import (
    parse_variable_expression_body,
    try_extract_variable_expression,
)


@dataclass
class Event:
    """An event represents some form of client/other criteria occurring (trigger). When an event trigger is met,
    any associated actions with the parent Step will be run.

    Events can have checks that must be all returned True/Pass at the moment the event is triggered otherwise
    the event trigger will be ignored."""

    type: str  # The type of event being listened for
    parameters: dict[str, Any]  # Any parameters to the event listener
    checks: list[Check] | None = None  # This event will be blocked from triggering if any of these checks return False

    def __post_init__(self):
        """Some parameter values might contain variable expressions (eg: a string "$now") that needs to be replaced
        with an parsed Expression object instead."""
        for k, v in self.parameters.items():
            variable_expr = try_extract_variable_expression(v)
            if variable_expr:
                self.parameters[k] = parse_variable_expression_body(variable_expr, k)


# The parameter schema for each event, keyed by the event name
EVENT_PARAMETER_SCHEMA: dict[str, dict[str, ParameterSchema]] = {
    "GET-request-received": {
        "endpoint": ParameterSchema(True, ParameterType.String),
        "serve_request_first": ParameterSchema(False, ParameterType.Boolean),
    },
    "POST-request-received": {
        "endpoint": ParameterSchema(True, ParameterType.String),
        "serve_request_first": ParameterSchema(False, ParameterType.Boolean),
    },
    "PUT-request-received": {
        "endpoint": ParameterSchema(True, ParameterType.String),
        "serve_request_first": ParameterSchema(False, ParameterType.Boolean),
    },
    "DELETE-request-received": {
        "endpoint": ParameterSchema(True, ParameterType.String),
        "serve_request_first": ParameterSchema(False, ParameterType.Boolean),
    },
    "wait": {"duration_seconds": ParameterSchema(True, ParameterType.Integer)},
}
VALID_EVENT_NAMES: set[str] = set(EVENT_PARAMETER_SCHEMA.keys())


def validate_event_parameters(procedure_name: str, step: str, event: Event) -> None:
    """Validates the event parameters for the parent TestProcedure based on the  EVENT_PARAMETER_SCHEMA

    raises TestProcedureDefinitionError on failure"""
    location = f"{procedure_name}.{step} Event: {event.type}"  # Descriptive location of this event being validated

    parameter_schema = EVENT_PARAMETER_SCHEMA.get(event.type, None)
    if parameter_schema is None:
        raise TestProcedureDefinitionError(f"{location} not a valid action name. {VALID_EVENT_NAMES}")

    validate_parameters(location, event.parameters, parameter_schema)



import pytest
from cactus_test_definitions.csipaus import (
    CSIPAusReadingLocation,
    CSIPAusReadingType,
    CSIPAusResource,
    is_list_resource,
)


@pytest.mark.parametrize("enum_val", CSIPAusResource)
def test_CSIPAusResource_name_matches_value(enum_val: CSIPAusResource):
    """We want case varying enum values (to match the values in CSIP Aus). This means we need to manage the names
    manually and ensures we don't end up with a bad copy paste"""
    assert enum_val.value == enum_val.name


@pytest.mark.parametrize("enum_val", CSIPAusReadingLocation)
def test_CSIPAusReadingLocation_name_matches_value(enum_val: CSIPAusReadingLocation):
    """We want case varying enum values (to match the values in CSIP Aus). This means we need to manage the names
    manually and ensures we don't end up with a bad copy paste"""
    assert enum_val.value == enum_val.name


@pytest.mark.parametrize("enum_val", CSIPAusReadingType)
def test_CSIPAusReadingType_name_matches_value(enum_val: CSIPAusReadingType):
    """We want case varying enum values (to match the values in CSIP Aus). This means we need to manage the names
    manually and ensures we don't end up with a bad copy paste"""
    assert enum_val.value == enum_val.name


@pytest.mark.parametrize(
    "resource, expected",
    [
        (CSIPAusResource.ConnectionPoint, False),
        (CSIPAusResource.DeviceCapability, False),
        (CSIPAusResource.DERSettings, False),
        (CSIPAusResource.EndDeviceList, True),
        (CSIPAusResource.EndDevice, False),
        (CSIPAusResource.DERList, True),
        (CSIPAusResource.DERProgramList, True),
        (CSIPAusResource.MirrorUsagePointList, True),
        (CSIPAusResource.DERControlList, True),
        (CSIPAusResource.DERControl, False),
    ],
)
def test_is_list_resource(resource: CSIPAusResource, expected: bool):
    actual = is_list_resource(resource)
    assert isinstance(actual, bool)
    assert actual is expected



from datetime import datetime, timezone
from decimal import Decimal
from itertools import product
from typing import Any

import pytest
from cactus_test_definitions.errors import TestProcedureDefinitionError
from cactus_test_definitions.parameters import (
    ParameterSchema,
    ParameterType,
    is_valid_parameter_type,
    validate_parameters,
)
from cactus_test_definitions.variable_expressions import (
    Constant,
    Expression,
    NamedVariable,
    NamedVariableType,
    OperationType,
)


@pytest.mark.parametrize(
    "value, type",
    product(
        [
            None,
            Expression(OperationType.ADD, Constant(1), Constant(2)),
            Constant(1),
            NamedVariable(NamedVariableType.NOW),
        ],
        list(ParameterType),
    ),
)
def test_is_valid_parameter_type_skipped_values(value: Any, type: ParameterType):
    """Tests that a number of values aren't considered and always return True"""
    actual = is_valid_parameter_type(type, value)
    assert isinstance(actual, bool)
    assert actual


@pytest.mark.parametrize(
    "type, value, expected",
    [
        (ParameterType.Boolean, True, True),
        (ParameterType.Boolean, False, True),
        (ParameterType.Boolean, 1, False),
        (ParameterType.Boolean, [1], False),
        (ParameterType.Boolean, "True", False),
        (ParameterType.Float, 0, True),
        (ParameterType.Float, 0.0, True),
        (ParameterType.Float, 1, True),
        (ParameterType.Float, 1.1, True),
        (ParameterType.Float, Decimal("1.1"), True),
        (ParameterType.Float, "1.1", False),
        (ParameterType.Float, [1.1], False),
        (ParameterType.Integer, 0, True),
        (ParameterType.Integer, 0.0, True),
        (ParameterType.Integer, 2, True),
        (ParameterType.Integer, 2.0, True),
        (ParameterType.Integer, Decimal("2.0"), True),
        (ParameterType.Integer, Decimal("2.1"), False),
        (ParameterType.Integer, 2.1, False),
        (ParameterType.Integer, [2], False),
        (ParameterType.Integer, "2", False),
        (ParameterType.DateTime, datetime(2022, 11, 1, 1, 2, 3), True),
        (ParameterType.DateTime, datetime(2022, 11, 1, 1, 2, 3, tzinfo=timezone.utc), True),
        (ParameterType.DateTime, "2022-11-01T01:02:03Z", False),
        (ParameterType.DateTime, "2022-11-01T01:02:03", False),
        (ParameterType.DateTime, 2022, False),
        (ParameterType.DateTime, [], False),
        (ParameterType.String, "", True),
        (ParameterType.String, "abc", True),
        (ParameterType.String, "12", True),
        (ParameterType.String, 12, False),
        (ParameterType.String, ["a"], False),
        (ParameterType.ListString, [], True),
        (ParameterType.ListString, [""], True),
        (ParameterType.ListString, ["", "b"], True),
        (ParameterType.ListString, ["", 4, "b"], False),
        (ParameterType.ListString, False, False),
        (ParameterType.ListString, True, False),
        (ParameterType.ListString, 123, False),
        (ParameterType.HexBinary, "AB", True),
        (ParameterType.HexBinary, "XY", False),
        (ParameterType.HexBinary, 12, False),
        (ParameterType.HexBinary, ["AB"], False),
        (ParameterType.HexBinary, 0.0, False),
        (ParameterType.HexBinary, [12], False),
        (ParameterType.HexBinary, 1.2, False),
        (ParameterType.HexBinary, Decimal("1.0"), False),
        (ParameterType.HexBinary, True, False),
        (ParameterType.CSIPAusResource, True, False),
        (ParameterType.CSIPAusResource, 123, False),
        (ParameterType.CSIPAusResource, "NotAResourceType", False),
        (ParameterType.CSIPAusResource, "ActivePowerInstantaneous", False),
        (ParameterType.CSIPAusResource, "EndDevice", True),
        (ParameterType.CSIPAusResource, ["EndDevice"], False),
        (ParameterType.ListCSIPAusResource, True, False),
        (ParameterType.ListCSIPAusResource, 123, False),
        (ParameterType.ListCSIPAusResource, "NotAResourceType", False),
        (ParameterType.ListCSIPAusResource, "EndDevice", False),
        (ParameterType.ListCSIPAusResource, ["EndDevice"], True),
        (ParameterType.ListCSIPAusResource, ["EndDevice", "NotAResource"], False),
        (ParameterType.ListCSIPAusResource, ["ActivePowerInstantaneous"], False),
        (ParameterType.CSIPAusReadingType, True, False),
        (ParameterType.CSIPAusReadingType, 123, False),
        (ParameterType.CSIPAusReadingType, "NotAResourceType", False),
        (ParameterType.CSIPAusReadingType, "EndDevice", False),
        (ParameterType.CSIPAusReadingType, "ActivePowerInstantaneous", True),
        (ParameterType.CSIPAusReadingType, ["ActivePowerSiteInstantaneous"], False),
        (ParameterType.ListCSIPAusReadingType, True, False),
        (ParameterType.ListCSIPAusReadingType, 123, False),
        (ParameterType.ListCSIPAusReadingType, "NotAResourceType", False),
        (ParameterType.ListCSIPAusReadingType, "EndDevice", False),
        (ParameterType.ListCSIPAusReadingType, "ActivePowerInstantaneous", False),
        (ParameterType.ListCSIPAusReadingType, ["ActivePowerInstantaneous"], True),
        (ParameterType.ListCSIPAusReadingType, ["ActivePowerInstantaneous", "NotAResource"], False),
        (ParameterType.ListCSIPAusReadingType, ["EndDevice"], False),
        (ParameterType.CSIPAusReadingLocation, True, False),
        (ParameterType.CSIPAusReadingLocation, 123, False),
        (ParameterType.CSIPAusReadingLocation, "NotAResourceType", False),
        (ParameterType.CSIPAusReadingLocation, "EndDevice", False),
        (ParameterType.CSIPAusReadingLocation, "Site", True),
        (ParameterType.CSIPAusReadingLocation, ["Site"], False),
        (ParameterType.ReadingTypeValues, 123, False),
        (ParameterType.ReadingTypeValues, "ActivePowerInstantaneous", False),
        (ParameterType.ReadingTypeValues, [123], False),
        (ParameterType.ReadingTypeValues, {}, False),
        (
            ParameterType.ReadingTypeValues,
            {"ActivePowerInstantaneous": [100, 1.23], "ActivePowerAverage": [200, -300]},
            True,
        ),
        (
            ParameterType.ReadingTypeValues,
            {"ActivePowerInstantaneous": [100, 200], "ActivePowerAverage": [200, 300, 400]},
            False,
        ),
        (
            ParameterType.ReadingTypeValues,
            {"ActivePowerInstantaneous": [100, 200], "ActivePowerAverage": [200, 300], "Foo": [100, 200]},
            False,
        ),
        (
            ParameterType.ReadingTypeValues,
            {"ActivePowerInstantaneous": [100, 200], "ActivePowerAverage": [200, "1.23"]},
            False,
        ),
        (
            ParameterType.ReadingTypeValues,
            {"ActivePowerInstantaneous": 100, "ActivePowerAverage": 200},
            False,
        ),
    ],
)
def test_is_valid_parameter_type(type: ParameterType, value: Any, expected: bool):
    actual = is_valid_parameter_type(type, value)
    assert isinstance(actual, bool)
    assert actual == expected


@pytest.mark.parametrize(
    "parameters, schema, is_valid",
    [
        ({}, {}, True),
        ({"foo": 1}, {}, False),
        ({"foo": 1}, {"foo": ParameterSchema(True, ParameterType.Integer)}, True),
        ({"foo": 1}, {"foo": ParameterSchema(True, ParameterType.Float)}, True),
        ({"foo": 1}, {"foo": ParameterSchema(True, ParameterType.String)}, False),  # wrong type
        (
            {"foo": 1},
            {"foo": ParameterSchema(True, ParameterType.Integer), "bar": ParameterSchema(False, ParameterType.String)},
            True,
        ),  # Optional param is OK to miss
        (
            {"foo": 1},
            {"foo": ParameterSchema(True, ParameterType.Integer), "bar": ParameterSchema(True, ParameterType.String)},
            False,
        ),  # Missing mandatory param
        (
            {"foo": None, "bar": "2", "baz": 4},
            {
                "foo": ParameterSchema(True, ParameterType.Integer),
                "bar": ParameterSchema(True, ParameterType.String),
                "baz": ParameterSchema(False, ParameterType.Float),
            },
            True,
        ),
    ],
)
def test_validate_action_parameters(parameters: dict, schema: dict, is_valid: bool):
    if is_valid:
        validate_parameters("foo", parameters, schema)
    else:
        with pytest.raises(TestProcedureDefinitionError):
            validate_parameters("foo", parameters, schema)



import pytest
from cactus_test_definitions.client.test_procedures import (
    TestProcedure,
    TestProcedureConfig,
    TestProcedureId,
)
from cactus_test_definitions.variable_expressions import (
    NamedVariableType,
    has_named_variable,
)

# Failures here will raise an issue in the test_from_yamlfile test
ALL_TEST_PROCEDURES: list[tuple[str, TestProcedure]] = []
try:
    ALL_TEST_PROCEDURES = [(k, tp) for k, tp in TestProcedureConfig.from_resource().test_procedures.items()]
except Exception:
    pass


def test_from_yamlfile():
    """This test confirms the standard test procedure yaml file (intended for production use)
    can be read and converted to the appropriate dataclasses.
    """
    test_procedures = TestProcedureConfig.from_resource()
    test_procedures.validate()


def test_TestProcedureId_synchronised():
    """Ensures that TestProcedureId is in sync with all available TestProcedures"""
    available_tests = set(TestProcedureConfig.available_tests())
    for t in available_tests:
        assert t in TestProcedureId, "TestProcedureConfig has a procedure not encoded in TestProcedureId"

    # By convention - test ALL-01 will be an enum ALL_01
    # for t in (t.replace("_", "-") for t in TestProcedureId):
    for t in TestProcedureId:
        assert t.value in available_tests, "TestProcedureId has extra procedures not found in TestProcedureConfig"


def test_available_tests_populated():
    """Force test procedures to load and ensure they all validate (and we at least have a few)"""
    assert len(TestProcedureConfig.available_tests()) > 0


@pytest.mark.parametrize("tp_id, tp", ALL_TEST_PROCEDURES)
def test_each_step_accessible(tp_id: str, tp: TestProcedure):
    """Ensures that each test procedure's steps are enabled/removed once. Failures here indicate a bad reference to
    enable/remove steps"""

    first_step = [k for k in tp.steps.keys()][0]
    step_names = set(tp.steps.keys())

    removal_count = dict([(s, 0) for s in step_names])
    enabled_count = dict([(s, 0) for s in step_names])

    # Tally up the number of times each step is added/removed via actions
    for s in tp.steps.values():
        for enabled_name in [name for a in s.actions if a.type == "enable-steps" for name in a.parameters["steps"]]:
            enabled_count[enabled_name] = enabled_count[enabled_name] + 1
        for removed_name in [name for a in s.actions if a.type == "remove-steps" for name in a.parameters["steps"]]:
            removal_count[removed_name] = removal_count[removed_name] + 1

    # Also check the preconditions for actions
    if tp.preconditions and tp.preconditions.actions:
        for enabled_name in [
            name for a in tp.preconditions.actions if a.type == "enable-steps" for name in a.parameters["steps"]
        ]:
            enabled_count[enabled_name] = enabled_count[enabled_name] + 1
        for removed_name in [
            name for a in tp.preconditions.actions if a.type == "remove-steps" for name in a.parameters["steps"]
        ]:
            removal_count[removed_name] = removal_count[removed_name] + 1

    # Each test should be added once and removed once
    for step_name in step_names:
        assert (
            removal_count[step_name] == 1
        ), f"{step_name}: Each test should be removed once otherwise the test cannot complete"

        if step_name == first_step:
            assert enabled_count[step_name] == 0, f"{step_name}: The first test step is already enabled"
        else:
            assert (
                enabled_count[step_name] == 1
            ), f"{step_name}: Each test should be added once - not expecting loops of steps"


@pytest.mark.parametrize("tp_id, tp", ALL_TEST_PROCEDURES)
def test_each_step_connected(tp_id: str, tp: TestProcedure):
    """Interprets each test a graph of steps linking to other steps (via enable-steps). Passes if the test procedure
    is fully connected from the first step (i.e. there are no isolated islands of steps that can't be reached from
    the first step)"""

    first_step = [k for k in tp.steps.keys()][0]
    step_names = set(tp.steps.keys())

    step_links_by_name: dict[str, list[str]] = {}
    for step_name in step_names:
        step = tp.steps[step_name]
        child_step_names = [name for a in step.actions if a.type == "enable-steps" for name in a.parameters["steps"]]
        step_links_by_name[step_name] = child_step_names

    # Walk through our simple graph/tree - make sure we visit every node in that tree
    def gather_connected_steps(links: dict[str, list[str]], node: str, visited_nodes: set[str]) -> None:
        if node in visited_nodes:
            return
        visited_nodes.add(node)
        for child in links[node]:
            gather_connected_steps(links, child, visited_nodes)

    visited_nodes: set[str] = set()
    gather_connected_steps(step_links_by_name, first_step, visited_nodes)

    # Also note that steps might be started as a precondition - so include those as visited too
    if tp.preconditions and tp.preconditions.actions:
        for n in [name for a in tp.preconditions.actions if a.type == "enable-steps" for name in a.parameters["steps"]]:
            visited_nodes.add(n)

    assert (
        step_names == visited_nodes
    ), "Missing entries here indicate a step (or steps) that can't be reached from the root node"


@pytest.mark.parametrize("_, tp", ALL_TEST_PROCEDURES)
def test_procedures_have_required_preconditions(_: str, tp: TestProcedure):

    # Immediate start implies that EndDevice registration will happen during the test body - these tests
    # don't require end-device-contents (as it will cause the test to fail anyway)
    is_immediate_start = tp.preconditions is not None and tp.preconditions.immediate_start
    if not is_immediate_start:
        assert tp.preconditions is not None, "Expected precondition check 'end-device-contents'"
        assert tp.preconditions.checks is not None, "Expected precondition check 'end-device-contents'"
        assert any(
            [check.type == "end-device-contents" for check in tp.preconditions.checks]
        ), "Expected precondition check 'end-device-contents'"

    # Check 'der-settings-contents' present if any precondition action parameter references setMaxW
    if tp.preconditions is not None and tp.preconditions.actions is not None:
        for action in tp.preconditions.actions:
            if action.parameters is not None:
                if any(
                    [
                        has_named_variable(
                            parameter_value=parameter, named_variable=NamedVariableType.DERSETTING_SET_MAX_W
                        )
                        for parameter in action.parameters.values()
                    ]
                ):
                    assert tp.preconditions.checks is not None
                    assert any([check.type == "der-settings-contents" for check in tp.preconditions.checks])



from cactus_test_definitions.client.actions import ACTION_PARAMETER_SCHEMA, Action
from cactus_test_definitions.client.checks import CHECK_PARAMETER_SCHEMA, Check
from cactus_test_definitions.client.events import EVENT_PARAMETER_SCHEMA, Event
from cactus_test_definitions.client.test_procedures import (
    Preconditions,
    Step,
    TestProcedure,
    TestProcedureConfig,
    TestProcedureId,
    TestProcedures,
)

__all__ = [
    "TestProcedureId",
    "Event",
    "Action",
    "ACTION_PARAMETER_SCHEMA",
    "Check",
    "CHECK_PARAMETER_SCHEMA",
    "EVENT_PARAMETER_SCHEMA",
    "Step",
    "Preconditions",
    "TestProcedure",
    "TestProcedures",
    "TestProcedureConfig",
]



import yaml


class UniqueKeyLoader(yaml.SafeLoader):
    """Originally sourced from https://gist.github.com/pypt/94d747fe5180851196eb
    Prevents duplicate keys from overwriting eachother instead of raising a ValueError.

    eg - consider the following YAML, it will parse OK but should be treated as an error:

    my_class:
        key1: abc
        key1: def
    """

    def construct_mapping(self, node, deep=False):
        mapping = set()
        for key_node, _ in node.value:
            key = self.construct_object(key_node, deep=deep)
            if key in mapping:
                raise ValueError(f"Duplicate {key!r} key found in YAML.")
            mapping.add(key)
        return super().construct_mapping(node, deep)



from dataclasses import dataclass
from typing import Any

from cactus_test_definitions.errors import TestProcedureDefinitionError
from cactus_test_definitions.parameters import (
    ParameterSchema,
    ParameterType,
    validate_parameters,
)
from cactus_test_definitions.variable_expressions import (
    parse_variable_expression_body,
    try_extract_variable_expression,
)


@dataclass
class Action:
    type: str
    parameters: dict[str, Any]

    def __post_init__(self):
        """Some parameter values might contain variable expressions (eg: a string "$now") that needs to be replaced
        with an parsed Expression object instead."""
        for k, v in self.parameters.items():
            variable_expr = try_extract_variable_expression(v)
            if variable_expr:
                self.parameters[k] = parse_variable_expression_body(variable_expr, k)


# The parameter schema for each action, keyed by the action name
ACTION_PARAMETER_SCHEMA: dict[str, dict[str, ParameterSchema]] = {
    "enable-steps": {"steps": ParameterSchema(True, ParameterType.ListString)},
    "remove-steps": {"steps": ParameterSchema(True, ParameterType.ListString)},
    "finish-test": {},
    "set-default-der-control": {
        "derp_id": ParameterSchema(False, ParameterType.Integer),
        "opModImpLimW": ParameterSchema(False, ParameterType.Float),
        "opModExpLimW": ParameterSchema(False, ParameterType.Float),
        "opModGenLimW": ParameterSchema(False, ParameterType.Float),
        "opModLoadLimW": ParameterSchema(False, ParameterType.Float),
        "opModStorageTargetW": ParameterSchema(False, ParameterType.Float),
        "setGradW": ParameterSchema(False, ParameterType.Integer),  # Hundredths of a percent / second
        "cancelled": ParameterSchema(False, ParameterType.Boolean),
    },
    "create-der-control": {
        "start": ParameterSchema(True, ParameterType.DateTime),
        "duration_seconds": ParameterSchema(True, ParameterType.Integer),
        "pow_10_multipliers": ParameterSchema(False, ParameterType.Integer),
        "primacy": ParameterSchema(False, ParameterType.Integer),
        "fsa_id": ParameterSchema(False, ParameterType.Integer),
        "randomizeStart_seconds": ParameterSchema(False, ParameterType.Integer),
        "ramp_time_seconds": ParameterSchema(False, ParameterType.Float),
        "opModEnergize": ParameterSchema(False, ParameterType.Boolean),
        "opModConnect": ParameterSchema(False, ParameterType.Boolean),
        "opModImpLimW": ParameterSchema(False, ParameterType.Float),
        "opModExpLimW": ParameterSchema(False, ParameterType.Float),
        "opModGenLimW": ParameterSchema(False, ParameterType.Float),
        "opModLoadLimW": ParameterSchema(False, ParameterType.Float),
        "opModFixedW": ParameterSchema(False, ParameterType.Float),
        "opModStorageTargetW": ParameterSchema(False, ParameterType.Float),
    },
    "create-der-program": {
        "primacy": ParameterSchema(True, ParameterType.Integer),
        "fsa_id": ParameterSchema(False, ParameterType.Integer),
    },
    "cancel-active-der-controls": {},
    "set-comms-rate": {
        "dcap_poll_seconds": ParameterSchema(False, ParameterType.Integer),
        "edev_post_seconds": ParameterSchema(False, ParameterType.Integer),
        "edev_list_poll_seconds": ParameterSchema(False, ParameterType.Integer),
        "fsa_list_poll_seconds": ParameterSchema(False, ParameterType.Integer),
        "derp_list_poll_seconds": ParameterSchema(False, ParameterType.Integer),
        "der_list_poll_seconds": ParameterSchema(False, ParameterType.Integer),
        "mup_post_seconds": ParameterSchema(False, ParameterType.Integer),
    },
    "communications-status": {"enabled": ParameterSchema(True, ParameterType.Boolean)},
    "edev-registration-links": {"enabled": ParameterSchema(True, ParameterType.Boolean)},
    "register-end-device": {
        "nmi": ParameterSchema(False, ParameterType.String),
        "registration_pin": ParameterSchema(False, ParameterType.Integer),
        "aggregator_lfdi": ParameterSchema(False, ParameterType.HexBinary),
        "aggregator_sfdi": ParameterSchema(False, ParameterType.Integer),
    },
}
VALID_ACTION_NAMES: set[str] = set(ACTION_PARAMETER_SCHEMA.keys())


def validate_action_parameters(procedure_name: str, step: str, action: Action) -> None:
    """Validates the action parameters for the parent TestProcedure based on the  ACTION_PARAMETER_SCHEMA

    raises TestProcedureDefinitionError on failure"""
    location = f"{procedure_name}.{step} Action: {action.type}"  # Descriptive location of this action being validated

    parameter_schema = ACTION_PARAMETER_SCHEMA.get(action.type, None)
    if parameter_schema is None:
        raise TestProcedureDefinitionError(f"{location} not a valid action name. {VALID_ACTION_NAMES}")

    validate_parameters(location, action.parameters, parameter_schema)



from datetime import datetime, timedelta, timezone
from pathlib import Path

import pytest
from cactus_test_definitions.client import TestProcedureConfig
from cactus_test_definitions.errors import TestProcedureDefinitionError
from cactus_test_definitions.variable_expressions import (
    Constant,
    Expression,
    NamedVariable,
    NamedVariableType,
    OperationType,
)


@pytest.mark.parametrize(
    "filename",
    [
        # No 'steps' parameters defined for enable-steps action (NOT-A-VALID-PARAMETER-NAME supplied instead)
        "tests/data/config_with_errors1.yaml",
        # Action refers to step "NOT-A-VALID-STEP"
        "tests/data/config_with_errors2.yaml",
    ],
)
def test_TestProcedures_validate_raises_exception(filename: str):

    with pytest.raises(TestProcedureDefinitionError):
        _ = TestProcedureConfig.from_yamlfile(path=Path(filename))


def test_TestProcedures_action_parameter_types():
    """Tests that lists/dicts/constants/datetimes can all be encoded/decoded via the yaml action definition"""
    cfg = TestProcedureConfig.from_yamlfile(path=Path("tests/data/config_action_parameters.yaml"), skip_validation=True)
    test_proc = cfg.test_procedures["CUSTOM-01"]
    step = test_proc.steps["Step-1"]
    action = step.actions[0]

    assert action.parameters == {
        "param_list_str": ["List Item 1", "List Item 2", "List Item 3"],
        "param_list_int": [11, 22, 33],
        "param_list_mixed": [
            11,
            2.2,
            "List Item 3",
            datetime(2020, 1, 2, 3, 4, 5, tzinfo=timezone.utc),
        ],
        "param_dict_str": {"key1": "value1", "key2": "value2"},
        "param_dict_int": {"key11": 11, "key22": 22},
        "param_int": 11,
        "param_str": "Value 11",
        "param_datetime_utc": datetime(2025, 1, 2, 3, 4, 5, tzinfo=timezone.utc),
        "param_datetime_naive": datetime(2025, 1, 2, 3, 4, 5, tzinfo=None),
        "param_with_variable_date": NamedVariable(NamedVariableType.NOW),
        "param_with_variable_db_lookup": NamedVariable(NamedVariableType.DERSETTING_SET_MAX_W),
        "param_with_variable_relative_db_lookup": Expression(
            OperationType.MULTIPLY, Constant(0.25), NamedVariable(NamedVariableType.DERSETTING_SET_MAX_W)
        ),
        "param_with_variable_relative_date": Expression(
            OperationType.SUBTRACT, NamedVariable(NamedVariableType.NOW), Constant(timedelta(minutes=5))
        ),
    }



import pytest
from cactus_test_definitions.server.actions import ACTION_PARAMETER_SCHEMA
from cactus_test_definitions.server.test_procedures import (
    TestProcedure,
    TestProcedureConfig,
    TestProcedureId,
)

# Failures here will raise an issue in the test_from_yamlfile test
ALL_TEST_PROCEDURES: list[tuple[str, TestProcedure]] = []
try:
    ALL_TEST_PROCEDURES = [(k, tp) for k, tp in TestProcedureConfig.from_resource().test_procedures.items()]
except Exception:
    pass


def test_ALL_TEST_PROCEDURES_parsed():
    assert len(ALL_TEST_PROCEDURES) > 0


def test_from_yamlfile():
    """This test confirms the standard test procedure yaml file (intended for production use)
    can be read and converted to the appropriate dataclasses.
    """
    test_procedures = TestProcedureConfig.from_resource()
    test_procedures.validate()


def test_TestProcedureId_synchronised():
    """Ensures that TestProcedureId is in sync with all available TestProcedures"""
    available_tests = set(TestProcedureConfig.available_tests())
    for t in available_tests:
        assert t in TestProcedureId, "TestProcedureConfig has a procedure not encoded in TestProcedureId"

    # By convention - test ALL-01 will be an enum ALL_01
    # for t in (t.replace("_", "-") for t in TestProcedureId):
    for t in TestProcedureId:
        assert t.value in available_tests, "TestProcedureId has extra procedures not found in TestProcedureConfig"


def test_available_tests_populated():
    """Force test procedures to load and ensure they all validate (and we at least have a few)"""
    assert len(TestProcedureConfig.available_tests()) > 0


@pytest.mark.parametrize("tp_id, tp", ALL_TEST_PROCEDURES)
def test_each_step_id_unique(tp_id: str, tp: TestProcedure):
    all_ids = [s.id for s in tp.steps]
    assert list(sorted(all_ids)) == list(sorted(set(all_ids)))
    assert len(set((s.id for s in tp.steps))) == len(tp.steps), "All steps must have a unique id property"


@pytest.mark.parametrize("tp_id, tp", ALL_TEST_PROCEDURES)
def test_each_alias_defined(tp_id: str, tp: TestProcedure):
    """Ensures that each test procedure's steps that have actions using an alias... define those aliases in advance"""

    # sanity check - make sure we are looking for the action names that are valid
    UPSERT_MUP_ACTION = "upsert-mup"
    INSERT_MUP_ACTION = "insert-readings"
    CREATE_SUB_ACTION = "create-subscription"
    DELETE_SUB_ACTION = "delete-subscription"
    assert UPSERT_MUP_ACTION in ACTION_PARAMETER_SCHEMA, "If this fails - the action name has changed. Update this test"
    assert INSERT_MUP_ACTION in ACTION_PARAMETER_SCHEMA, "If this fails - the action name has changed. Update this test"
    assert CREATE_SUB_ACTION in ACTION_PARAMETER_SCHEMA, "If this fails - the action name has changed. Update this test"
    assert DELETE_SUB_ACTION in ACTION_PARAMETER_SCHEMA, "If this fails - the action name has changed. Update this test"

    mup_aliases_found: set[str] = set()
    sub_aliases_found: set[str] = set()
    for step in tp.steps:
        if step.action.parameters:
            mup_id = step.action.parameters.get("mup_id", None)
            sub_id = step.action.parameters.get("sub_id", None)
        else:
            mup_id = None
            sub_id = None

        if step.action.type == UPSERT_MUP_ACTION:
            assert mup_id and (mup_id not in mup_aliases_found), "mup_id is already defined in this test."
            mup_aliases_found.add(mup_id)
        elif step.action.type == INSERT_MUP_ACTION:
            assert mup_id and (mup_id in mup_aliases_found), "mup_id hasn't been defined yet."
        elif step.action.type == CREATE_SUB_ACTION:
            assert sub_id and (sub_id not in sub_aliases_found), "sub_id is already defined in this test."
            sub_aliases_found.add(sub_id)
        elif step.action.type == DELETE_SUB_ACTION:
            assert sub_id and (sub_id in sub_aliases_found), "sub_id hasn't been defined yet."
